@@ -5,32 +5,22 @@ from dataclasses import dataclass
 import logging
 import random
 import time
+from enum import Enum
 from collections import defaultdict
 
 from .discovery import P2PDiscovery
 from .filesharing import pack_job
-from .sync_objects import CPReplDict, CPReplLockManager
+from .sync_objects import CPOrderedReplDict, CPReplLockManager
 
-
-# Peer dict is keyed by the peer addr. Value is tuple(last_update_ts, state)
-# where state is an opaque dict.
-class PeerDict(CPReplDict):
-    def _item_changed(self, prev, nxt):
-        if prev is None:
-            return True
-        for k in ('status', 'run'):
-            if prev[1].get(k) != nxt[1].get(k):
-                return True
-        return False
-
-# Job dict is keyed by the hash of the .gjob file. Value is tuple(submitting_peer_addr, manifest)
-# where manifest is an opaque dict.
-class JobDict(CPReplDict):
-    def _item_changed(self, prev, nxt):
-        return True # always trigger callback 
+class ChangeType(Enum):
+    PEER = "peer"
+    JOB = "job"
+    LOCK = "lock"
+    QUEUE = "queue"
 
 # This queue is shared with other printers on the local network which are configured with the same namespace.
 # Actual scheduling and printing is done by the object owner.
+# Job data should be considered opaque at this level - 
 class LANPrintQueueBase():
     PEER_TIMEOUT = 60
 
@@ -43,10 +33,26 @@ class LANPrintQueueBase():
       self.update_cb = update_cb 
       self._syncobj = None
 
+    def _tagged_cb(self, changetype):
+        def tagcb(prev, nxt):
+            # Unwrap metadata tuples and return end value (user provided)
+            if prev is not None and type(prev) is tuple:
+                prev = prev[-1]
+            if nxt is not None and type(nxt) is tuple:
+                nxt = nxt[-1]
+            self.update_cb(changetype, prev, nxt)
+        return tagcb
+
     def connect(self, peers):
-        self.peers = PeerDict(self.update_cb)
-        self.jobs = JobDict(self.update_cb)
-        self.locks = CPReplLockManager(selfID=self.addr, autoUnlockTime=60, cb=self.update_cb)
+        # Peer dict is keyed by the peer addr. Value is tuple(last_update_ts, state)
+        # where state is an opaque dict.
+        self.peers = CPOrderedReplDict(self._tagged_cb(ChangeType.PEER))
+
+        # Job dict is keyed by the ID of the job. Value is tuple(submitting_peer_addr, manifest)
+        # where manifest is an opaque dict.
+        self.jobs = CPOrderedReplDict(self._tagged_cb(ChangeType.JOB))
+
+        self.locks = CPReplLockManager(selfID=self.addr, autoUnlockTime=60, cb=self._tagged_cb(ChangeType.LOCK))
         conf = SyncObjConf(
                 onReady=self.on_ready, 
                 dynamicMembershipChange=True,
@@ -60,7 +66,7 @@ class LANPrintQueueBase():
         # Set ready state on all objects, enabling callbacks now
         # that they've been fast-forwarded
         self._logger.info("LANPrintQueueBase.on_ready")
-        self.update_cb()
+        self.update_cb(ChangeType.QUEUE, False, True)
 
     # ==== Network methods ====
 
@@ -96,44 +102,48 @@ class LANPrintQueueBase():
       result = {}
       peerlocks = self.locks.getPeerLocks()
       for k, v in self.peers.items():
-          result[k] = dict(**v[1], acquired=peerlocks.get(k, []))
+          result[k] = dict(**v[1]) # Exclude peer update timestamp
       return result
 
-    def setJob(self, hash_, manifest: dict, addr=None):
+    def getPeer(self, peer):
+        p = self.peers.get(peer)
+        if p is not None:
+            return dict(**p[-1])
+
+    def hasJob(self, jid) -> bool:
+        return (jid in self.jobs)
+
+    def setJob(self, jid, manifest: dict, addr=None):
       # performed synchronously to prevent race conditions when quickly
       # writing, then reading job information
       if addr == None:
           addr = self.addr
-      self.jobs.set(hash_, (addr, manifest), sync=True, timeout=5.0)
+      self.jobs.set(jid, (addr, manifest), sync=True, timeout=5.0)
 
-    def getJobs(self):
-        jobs = []
+    def getLocks(self):
         joblocks = {}
         for (peer, locks) in self.locks.getPeerLocks().items():
             for lock in locks:
                 joblocks[lock] = peer
-        for (hash_, v) in self.jobs.items():
-            (peer, manifest) = v
-            job = dict(**manifest, peer_=peer, acquired_by_=joblocks.get(hash_))
-            # Ensure IDs are up to date
-            job['id'] = hash_
-            for i, s in enumerate(job['sets']):
-                s['id'] = f"{hash_}_{i}"
-            jobs.append(job)
-        # Note that jobs are returned unordered; caller can sort it after the fact.
-        return jobs
+        return joblocks
 
-    def removeJob(self, hash_: str):
-      self.jobs.pop(hash_, None)
+    def getJobs(self):
+        return self.jobs.ordered_items()
 
-    def acquireJob(self, hash_: str):
-      try:
-        return self.locks.tryAcquire(hash_, sync=True, timeout=self.acquire_timeout)
-      except SyncObjException: # timeout
-        return False
+    def getJob(self, jid):
+        return self.jobs.get(jid)
 
-    def releaseJob(self, hash_: str):
-      self.locks.release(hash_)
+    def removeJob(self, jid: str):
+        return self.jobs.pop(jid, None)
+
+    def acquireJob(self, jid: str):
+        try:
+            return self.locks.tryAcquire(jid, sync=True, timeout=self.acquire_timeout)
+        except SyncObjException: # timeout
+            return False
+
+    def releaseJob(self, jid: str):
+        self.locks.release(jid)
 
 
 # Wrap LANPrintQueueBase in a discovery class, allowing for dynamic membership based 
@@ -172,11 +182,11 @@ class LANPrintQueue(P2PDiscovery):
       self._logger.info(f"Host removed: {host}")
       self.q.removePeer(host)
 
-  def _init_base(self, results):
+  def _init_base(self):
     self.q = LANPrintQueueBase(self.ns, self.addr, self.update_cb, self._logger)
 
   def _on_startup_complete(self, results):
     self._logger.info(f"Discover end: {results}; initializing queue")
-    self._init_base(results)
+    self._init_base()
     self.q.connect(results.keys())
 
